#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/*
 * parse fib rule messages
 */
int parse_frhdr(struct nlmsghdr *nlh)
{
	struct fib_rule_hdr *frh = (struct fib_rule_hdr *)NLMSG_DATA(nlh);
	struct rtattr *fra[__FRA_MAX];
	char msg_type[16] = "", msg[128] = "", action[16] = "";
	char *mp = msg;

	/* check address family */
	if(frh->family == AF_INET)
		strcpy(msg_type, "IPv4 RULE");
	else if(frh->family == AF_INET6)
		strcpy(msg_type, "IPv6 RULE");
	else {
		rec_log("[%-12s] unknown address family: %d", "ERROR", frh->family);
		return(1);
	}

	/* check message type */
	if(nlh->nlmsg_type == RTM_NEWRULE)
		mp += sprintf(mp, "[%-12s] rule added: ", msg_type);
	else if(nlh->nlmsg_type == RTM_DELRULE)
		mp += sprintf(mp, "[%-12s] rule deleted: ", msg_type);
	else {
		rec_log("[%-12s] unknown netlink message type: %d", "ERROR", nlh->nlmsg_type);
		return(1);
	}

	/* check route table id(other than local) */
	if(frh->table == RT_TABLE_LOCAL)
		return(1);

	/* parse fib rule header */
	parse_frule(fra, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_frhdr(frh, fra);
	}

	/* check whether invert rule or not */
	if(frh->flags & FIB_RULE_INVERT)
		mp += sprintf(mp, "not ");

        /* get source prefix */
	if(fra[FRA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";
		inet_ntop(frh->family, RTA_DATA(fra[FRA_SRC]), src, sizeof(src));
		mp += sprintf(mp, "from=%s/%d ", src, frh->src_len);
	} 

	/* get egress interface index & name */
	if(fra[FRA_IIFNAME]) {
		char iifname[IFNAMSIZ]; 
		memcpy(iifname, RTA_DATA(fra[FRA_IIFNAME]), sizeof(iifname));
		if(frh->flags & FIB_RULE_IIF_DETACHED)
			mp += sprintf(mp, "in=%s(detached) ", iifname);
		else
			mp += sprintf(mp, "in=%s ", iifname);
	}

	/* get tos value */
	if(frh->tos)
		mp += sprintf(mp, "tos=0x%.2x ", frh->tos);

        /* get destination prefix */
	if(fra[FRA_DST]) {
		char dst[INET6_ADDRSTRLEN] = "";
		inet_ntop(frh->family, RTA_DATA(fra[FRA_DST]), dst, sizeof(dst));
		mp += sprintf(mp, "to=%s/%d ", dst, frh->dst_len);
	}

	/* get egress interface index & name */
	if(fra[FRA_OIFNAME]) {
		char oifname[IFNAMSIZ]; 
		memcpy(oifname, RTA_DATA(fra[FRA_OIFNAME]), sizeof(oifname));
		if(frh->flags & FIB_RULE_OIF_DETACHED)
			mp += sprintf(mp, "out=%s(detached) ", oifname);
		else
			mp += sprintf(mp, "out=%s ", oifname);
	}

	/* get fwmark and fwmask */
	if(fra[FRA_FWMARK] || fra[FRA_FWMASK]) {
		unsigned fwmark = 0, fwmask = 0;
		if(fra[FRA_FWMARK])
			fwmark = *(unsigned *)RTA_DATA(fra[FRA_FWMARK]);

		if(fra[FRA_FWMASK] &&
			(fwmask = *(unsigned *)RTA_DATA(fra[FRA_FWMASK])) != 0xFFFFFFFF)
			mp += sprintf(mp, "fwmark=0x%x/0x%x ", fwmark, fwmask);
		else
			mp += sprintf(mp, "fwmark=0x%x ", fwmark);
	}

	/* get source and destination realms */
	if(fra[FRA_FLOW]) {
		unsigned srcrlm = *(unsigned *)RTA_DATA(fra[FRA_FLOW]);
		unsigned dstrlm = srcrlm >> 16;
		srcrlm &= 0xFFFF;
		if(srcrlm)
			mp += sprintf(mp, "src_realm=%u ", srcrlm);
		mp += sprintf(mp, "dst_realm=%u ", dstrlm);
	}

	/* get another rule to junmp to */
	if(fra[FRA_GOTO])
		mp += sprintf(mp, "goto=%u ", *(unsigned *)RTA_DATA(fra[FRA_GOTO]));

	/* convert route table name */
	//rt_table_n2s(frh->table, table);
	if(fra[FRA_TABLE]) {
		char table[16] = "";
		rt_table_n2s(*(unsigned *)RTA_DATA(fra[FRA_TABLE]), table);
		if(strcmp(table, "none"))
			mp += sprintf(mp, "table=%s ", table);
	}

	/* get fib rule priority */
	if(fra[FRA_PRIORITY])
		mp += sprintf(mp, "priority=%u ", *(unsigned *)RTA_DATA(fra[FRA_PRIORITY]));

	/* convert fib rule action */
	fr_act_n2s(frh->action, action);

	/* logging fib rule */
	if(frh->flags & FIB_RULE_UNRESOLVED)
		rec_log("%saction=%s(unresolved)", msg, action);
	else
		rec_log("%saction=%s", msg, action);

	return(0);
}

/*
 * debugging fib rule messages
 */
void debug_frhdr(struct fib_rule_hdr *frh, struct rtattr *fra[])
{
	char str[16] = "";

	/* debug frhdr */
	rec_dbg("******** frhdr ******************************************************");
	rec_dbg("fib_rule_hdr:");
	rec_dbg("    family: %d(%s)", frh->family, af_type_n2s(frh->family));
	rec_dbg("    dst_len: %d", frh->dst_len);
	rec_dbg("    src_len: %d", frh->src_len);
	rec_dbg("    tos: 0x%.2x", frh->tos);
	rec_dbg("    table: %d", frh->table);
	rec_dbg("    res1: %d", frh->res1);
	rec_dbg("    res2: %d", frh->res2);

	fr_act_n2s(frh->action, str);
	rec_dbg("    action: %d(%s)", frh->action, str);

	unsigned char flags = frh->flags;
	char flist[36] = "";
#define _FIB_RULE(f) \
	if(flags & FIB_RULE_##f) \
		(flags &= ~FIB_RULE_##f) ? strcat(flist, #f ",") : strcat(flist, #f); 
	_FIB_RULE(PERMANENT);
	_FIB_RULE(INVERT);
	_FIB_RULE(UNRESOLVED);
	_FIB_RULE(IIF_DETACHED);
	_FIB_RULE(OIF_DETACHED);
#undef _FIB_RULE
	rec_dbg("    flags: %u(%s)", frh->flags, flist);

	/* debug fib rule attributes */
	if(fra[FRA_DST]) {
		char dst[INET6_ADDRSTRLEN] = "";
		inet_ntop(frh->family, RTA_DATA(fra[FRA_DST]), dst, sizeof(dst));
		rec_dbg("FRA_DST(%hu): %s", fra[FRA_DST]->rta_len, dst);
	}

	if(fra[FRA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";
		inet_ntop(frh->family, RTA_DATA(fra[FRA_SRC]), src, sizeof(src));
		rec_dbg("FRA_SRC(%hu): %s", fra[FRA_SRC]->rta_len, src);
	}

	if(fra[FRA_IIFNAME]) {
		char iifname[IFNAMSIZ]; 
		memcpy(iifname, RTA_DATA(fra[FRA_IIFNAME]), sizeof(iifname));
		rec_dbg("FRA_IIFNAME(%hu): %s", fra[FRA_IIFNAME]->rta_len, iifname);
	}

	/* 
	 * FRA_GOTO was committed in kernel 2.6.22.
	 * commit 947c9fe56d9cf7ad0bc3a03ccd30446cde698e4
	 * [NET] fib_rules: goto rule action
	 */
	if(fra[FRA_GOTO])
		rec_dbg("FRA_GOTO(%hu): %u", fra[FRA_GOTO]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_GOTO]));
	
	if(fra[FRA_PRIORITY])
		rec_dbg("FRA_PRIORITY(%hu): %u", fra[FRA_PRIORITY]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_PRIORITY]));

	if(fra[FRA_FWMARK])
		rec_dbg("FRA_FWMARK(%hu): %u", fra[FRA_FWMARK]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_FWMARK]));

	if(fra[FRA_FLOW])
		rec_dbg("FRA_FLOW(%hu): 0x%x", fra[FRA_FLOW]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_FLOW]));

	if(fra[FRA_TABLE])
		rec_dbg("FRA_TABLE(%hu): 0x%x", fra[FRA_TABLE]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_TABLE]));

	if(fra[FRA_FWMASK])
		rec_dbg("FRA_FWMASK(%hu): %u", fra[FRA_FWMASK]->rta_len, *(unsigned *)RTA_DATA(fra[FRA_FWMASK]));

	if(fra[FRA_OIFNAME]) {
		char oifname[IFNAMSIZ]; 
		memcpy(oifname, RTA_DATA(fra[FRA_OIFNAME]), sizeof(oifname));
		rec_dbg("FRA_OIFNAME(%hu): %s", fra[FRA_OIFNAME]->rta_len, oifname);
	}

	rec_dbg("");
}

/*
 * convert FR_ACT_* from number to string
 */
void fr_act_n2s(unsigned char nact, char *sact)
{
	switch(nact) {
		case FR_ACT_UNSPEC:
			strcpy(sact, "none");
			break;
		case FR_ACT_TO_TBL:
			strcpy(sact, "to_tbl");
			break;
		case FR_ACT_GOTO:
			strcpy(sact, "goto");
			break;
		case FR_ACT_NOP:
			strcpy(sact, "nop");
			break;
		case FR_ACT_RES3:
			strcpy(sact, "res3");
			break;
		case FR_ACT_RES4:
			strcpy(sact, "res4");
			break;
		case FR_ACT_BLACKHOLE:
			strcpy(sact, "blackhole");
			break;
		case FR_ACT_UNREACHABLE:
			strcpy(sact, "unreachable");
			break;
		case FR_ACT_PROHIBIT:
			strcpy(sact, "prohibit");
			break;
		default:
			strcpy(sact, "unknown");
			break;
	}
}
