#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/*
 * parse interface address messages
 */ 
int parse_ifamsg(struct nlmsghdr *nlh)
{
	struct ifaddrmsg *ifam = (struct ifaddrmsg *)NLMSG_DATA(nlh);
	struct rtattr *ifa[__IFA_MAX];
	char paddr[INET6_ADDRSTRLEN] = "";
	char laddr[INET6_ADDRSTRLEN] = "";
	char ifname[IFNAMSIZ] = "";
	char iflabel[IFNAMSIZ] = "";
	char msg[128] = "";
	char *mp = msg;
	int nmask = -1;
	unsigned short iftype = -1;

	/* check address family & set message type */
	if(ifam->ifa_family == AF_INET)
		mp += sprintf(mp, "[%-12s] ", "IPv4 ADDRESS");
	else if(ifam->ifa_family == AF_INET6)
		mp += sprintf(mp, "[%-12s] ", "IPv6 ADDRESS");
	else {
		rec_log("[%-12s] unknown address family: %d", "ERROR", ifam->ifa_family);
		return(1);
	}

	/* get interface index & name */
	if(nlh->nlmsg_type == RTM_NEWADDR) {
		if_indextoname(ifam->ifa_index, ifname);
		mp += sprintf(mp, "address added: ");
	} else if(nlh->nlmsg_type == RTM_DELADDR) {
		if(!if_indextoname_from_iflist(ifam->ifa_index, ifname))
			if_indextoname_from_ifhist(ifam->ifa_index, ifname);
		mp += sprintf(mp, "address deleted: ");
	} else {
		rec_log("[%-12s] unknown nlmsg_type: %d", "ERROR", nlh->nlmsg_type);
		return(1);
	}

	/* parse interface address attributes */
	parse_ifaddr(ifa, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ifamsg(ifam, ifa);
	}

	/* get prefix address */
	if(ifa[IFA_ADDRESS])
		inet_ntop(ifam->ifa_family, RTA_DATA(ifa[IFA_ADDRESS]), paddr, sizeof(paddr));

	/* get local interface address */
	if(ifa[IFA_LOCAL])
		inet_ntop(ifam->ifa_family, RTA_DATA(ifa[IFA_LOCAL]), laddr, sizeof(laddr));
	
	/* get interface type */
	iftype = get_type_from_iflist(ifam->ifa_index);
	if(!iftype)
		iftype = get_type_from_ifhist(ifam->ifa_index);

	/* whether interface type is P2P */
	if(iftype == ARPHRD_PPP)
		mp += sprintf(mp, "if=%s local=%s/%d peer=%s/%d",
			ifname, laddr, ifam->ifa_prefixlen, paddr, ifam->ifa_prefixlen);
	else
		mp += sprintf(mp, "if=%s ipaddr=%s/%d", ifname, paddr, ifam->ifa_prefixlen);

	/* get scope of address */
	switch(ifam->ifa_scope) {
		case RT_SCOPE_UNIVERSE:
			mp += sprintf(mp, " scope=global");
			break;
		case RT_SCOPE_SITE:
			mp += sprintf(mp, " scope=site");
			break;
		case RT_SCOPE_LINK:
			mp += sprintf(mp, " scope=link");
			break;
		case RT_SCOPE_HOST:
			mp += sprintf(mp, " scope=host");
			break;
		case RT_SCOPE_NOWHERE:
			mp += sprintf(mp, " scope=none");
			break;
		default:
			mp += sprintf(mp, " scope=unknown");
	}

	/* get interface label */
	if(ifa[IFA_LABEL]) {
		memcpy(iflabel, RTA_DATA(ifa[IFA_LABEL]), IFNAMSIZ);
		if(strcmp(iflabel, ifname))
			mp += sprintf(mp, " label=%s", iflabel);
	}

	/* logging interface address message */
	rec_log("%s", msg);

	return(0);
}

/*
 * debug interface address messages
 */ 
void debug_ifamsg(struct ifaddrmsg *ifam, struct rtattr *ifa[])
{
	unsigned char flags = ifam->ifa_flags;
	char flist[128] = "";

#define _IFA_F(f) \
	if(flags & IFA_F_##f) \
		(flags &= ~IFA_F_##f) ? strcat(flist, #f ",") : strcat(flist, #f);
	_IFA_F(SECONDARY);
	_IFA_F(TEMPORARY);
	_IFA_F(NODAD);
	_IFA_F(OPTIMISTIC);
	_IFA_F(DADFAILED);
	_IFA_F(HOMEADDRESS);
	_IFA_F(DEPRECATED);
	_IFA_F(TENTATIVE);
	_IFA_F(PERMANENT);
#undef _IFA_F

	char ifname[IFNAMSIZ] = "";
	if_indextoname(ifam->ifa_index, ifname);

	/* debug ifaddrmsg */
	rec_dbg("******** ifaddrmsg **************************************************");
	rec_dbg("ifaddrmsg:");
	rec_dbg("    ifa_family: %d(%s)", ifam->ifa_family, af_type_n2s(ifam->ifa_family));
	rec_dbg("    ifa_prefixlen: %d", ifam->ifa_prefixlen);
	rec_dbg("    ifa_flags: 0x%x(%s)", ifam->ifa_flags, flist);

#define _IFA_SCOPE(s) \
	if(ifam->ifa_scope == RT_SCOPE_##s) \
		rec_dbg("    ifa_scope: %d(%s)", ifam->ifa_scope, #s);
	_IFA_SCOPE(UNIVERSE);
	_IFA_SCOPE(SITE);
	_IFA_SCOPE(LINK);
	_IFA_SCOPE(HOST);
	_IFA_SCOPE(NOWHERE);
#undef _RT_SCOPE

	rec_dbg("    ifa_index: %u(%s)", ifam->ifa_index, ifname);

	/* debug interface address attributes */
	rec_dbg("******** attributes *************************************************");

	/* debug prefix address */
	if(ifa[IFA_ADDRESS]) {
		char paddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_ADDRESS]), paddr, sizeof(paddr));
		rec_dbg("IFA_ADDRESS(%hu): %s", ifa[IFA_ADDRESS]->rta_len, paddr);
	}

	/* debug local address */
	if(ifa[IFA_LOCAL]) {
		char laddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_LOCAL]), laddr, sizeof(laddr));
		rec_dbg("IFA_LOCAL(%hu): %s", ifa[IFA_LOCAL]->rta_len, laddr);
	}

	/* debug interface label */
	if(ifa[IFA_LABEL])
		rec_dbg("IFA_LABEL(%hu): %s", ifa[IFA_LABEL]->rta_len, (char *)RTA_DATA(ifa[IFA_LABEL]));

	/* debug interface broadcast address */
	if(ifa[IFA_BROADCAST]) {
		char baddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_BROADCAST]), baddr, sizeof(baddr));
		rec_dbg("IFA_BROADCAST(%hu): %s", ifa[IFA_BROADCAST]->rta_len, baddr);
	}

	/* debug interface anycast address */
	if(ifa[IFA_ANYCAST]) {
		char aaddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_ANYCAST]), aaddr, sizeof(aaddr));
		rec_dbg("IFA_ANYCAST(%hu): %s", ifa[IFA_ANYCAST]->rta_len, aaddr);
	}

	/* debug interface cache information */
	if(ifa[IFA_CACHEINFO]) {
		struct ifa_cacheinfo *ifac = (struct ifa_cacheinfo *)RTA_DATA(ifa[IFA_CACHEINFO]);
		rec_dbg("IFA_CACHEINFO(%hu):", ifa[IFA_CACHEINFO]->rta_len);
		rec_dbg("    ifa_cacheinfo:");
		rec_dbg("        ifa_prefered: 0x%x", ifac->ifa_prefered);
		rec_dbg("        ifa_valid: 0x%x", ifac->ifa_valid);
		rec_dbg("        cstamp: %u", ifac->cstamp);
		rec_dbg("        tstamp: %u", ifac->tstamp);
	}
	
	/* debug interface multicast address */
	if(ifa[IFA_MULTICAST]) {
		char maddr[INET6_ADDRSTRLEN] = "";
		inet_ntop(ifam->ifa_family,  RTA_DATA(ifa[IFA_MULTICAST]), maddr, sizeof(maddr));
		rec_dbg("IFA_ANYCAST(%hu): %s", ifa[IFA_MULTICAST]->rta_len, maddr);
	}

	rec_dbg("");

	return;
}
