#include "nield.h"
#include "rtnetlink.h"

#define HIST_MAX 8

extern int optflag;

/* top of interface list and history */
static LIST_HEAD(lhead);
static LIST_HEAD(hhead);

/*
 * create an interface list
 */ 
int create_iflist(struct msghdr *msg)
{
	struct nlmsghdr *nlh;
	int nlh_len;
	struct ifinfomsg *ifim;
	struct rtattr *ifla[__IFLA_MAX];
	struct rtattr *linkinfo[__IFLA_INFO_MAX];
	struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
	unsigned char *addr;
	struct iflist_entry *ifle;

	/* get netlink message header */
	nlh = msg->msg_iov->iov_base;
	nlh_len = msg->msg_iov->iov_len;

	/* parse netlink message header */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		/* whether netlink message header ends or not */
		if(nlh->nlmsg_type == NLMSG_DONE)
			return(1);

		/* get interface infomation message */
		ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);

		/* check protocol family(PF_UNSPEC only) */
		if(ifim->ifi_family != PF_UNSPEC)
			continue;

		/* parse interface infomation attributes */
		parse_ifinfo(ifla, nlh);

		/* create interface list */
		ifle = (struct iflist_entry *)malloc(sizeof(struct iflist_entry));
		if(ifle == NULL) {
			rec_log("[%-12s] malloc(): failed", "ERROR");
			return(-1);
		}
		memset(ifle, 0, sizeof(struct iflist_entry));
		INIT_LIST_HEAD(&(ifle->list));

		/* get interface name */
		ifle->index = ifim->ifi_index;
		if(ifla[IFLA_IFNAME])
			memcpy(ifle->name, RTA_DATA(ifla[IFLA_IFNAME]), sizeof(ifle->name));

		/* get interface address */
		if(ifla[IFLA_ADDRESS]) {
			char addrstr[INET6_ADDRSTRLEN+1] = "";
			if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
				RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
			memcpy(ifle->addr, addrstr, sizeof(addrstr));
		}

		/* get broadcast address */
		if(ifla[IFLA_BROADCAST]) {
			char brdstr[INET6_ADDRSTRLEN+1] = "";
			if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
				RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
			memcpy(ifle->brd, brdstr, sizeof(brdstr));
		}

		/* get interface broadcast address */
		if(ifla[IFLA_BROADCAST])
			memcpy(ifle->brd, RTA_DATA(ifla[IFLA_BROADCAST]), sizeof(ifle->brd));

		ifle->flags = ifim->ifi_flags;
		ifle->type = ifim->ifi_type;
		ifle->vid = 0;

		/* get interface MTU */
		if(ifla[IFLA_MTU])
			ifle->mtu = *(unsigned short *)RTA_DATA(ifla[IFLA_MTU]);

		/* debug */
		if(optflag & F_DEBUG) {
			debug_nlmsg(nlh);
			debug_ifimsg(ifim, ifla);
		}

		/* add interface list */
		list_add(&(ifle->list), &lhead);
	}

	return(0);
}

/*
 * print an interface list
 */ 
void print_iflist(void)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);

		printf("%d ", e->index);
		printf("IFLA_IFNAME: %s, ", e->name);
		printf("IFLA_ADDRESS: %s, ", e->addr);
		printf("IFLA_BROADCAST: %s, ", e->brd);
		printf("IFLA_MTU: %hu\n", e->mtu);
	}
}

/*
 * search an entry in an interface list
 */
static inline struct iflist_entry *search_iflist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * add an entry into an interface list
 */ 
static inline void add_iflist_entry(struct list_head *l)
{
	INIT_LIST_HEAD(l);
	list_add(l, &lhead);
}

/*
 * [not in use]
 * delete an entry in an interface list
 */ 
static inline void del_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_del(l);
			free(e);
		}
	}
}

/*
 * move an entry in an interface list
 */
static inline void move_iflist_entry(unsigned index)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &lhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index) {
			list_move(l, &hhead);
			break;
		}
	}
}

/*
 * delete an entry in an interface history
 */
static inline void del_ifhist_entry(void)
{
	struct list_head *l, *n;
	int i = 0;

	list_for_each_safe(l, n, &hhead) {
		struct iflist_entry *e;

		if(++i > HIST_MAX) {
			e = list_entry(l, struct iflist_entry, list);
			list_del(l);
			free(e);
			break;
		}
	}
}

/*
 * search an entry in an interface history
 */ 
static inline struct iflist_entry *search_ifhist(unsigned index)
{
	struct list_head *l;

	list_for_each(l, &hhead) {
		struct iflist_entry *e;

		e = list_entry(l, struct iflist_entry, list);
		if(e->index == index)
			return(e);
	}

	return(NULL);
}

/*
 * convert an interface index to name in an interface list
 */
char *if_indextoname_from_iflist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * convert an interface index to name in an inetrface history
 */ 
char *if_indextoname_from_ifhist(unsigned index, char *ifname)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(NULL);

	strncpy(ifname, e->name, sizeof(e->name));

	return(ifname);
}

/*
 * get an interface type in an interface list
 */ 
unsigned short get_type_from_iflist(unsigned index)
{
	struct iflist_entry *e = search_iflist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * get an interface type in an interface history
 */
unsigned short get_type_from_ifhist(unsigned index)
{
	struct iflist_entry *e = search_ifhist(index);

	if(!e)
		return(0);

	return(e->type);
}

/*
 * parse interface information messages
 */
int parse_ifimsg(struct nlmsghdr *nlh)
{
	struct ifinfomsg *ifim = (struct ifinfomsg *)NLMSG_DATA(nlh);
	struct rtattr *ifla[__IFLA_MAX];
	struct rtattr *linkinfo[__IFLA_INFO_MAX];
	struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
	struct iflist_entry *ifle_tmp, *ifle;
	char msg_type[16] = "", msg[128] = "";
	char *mp = msg;

	/* set message type */
	strcpy(msg_type, "INTERFACE");

	/* create new interface list entry */
	ifle_tmp = malloc(sizeof(struct iflist_entry));
	if(!ifle_tmp) {
		rec_log("[%-12s] malloc() failed", "ERROR");
		return(1);
	}
	memset(ifle_tmp, 0, sizeof(struct iflist_entry));
	INIT_LIST_HEAD(&(ifle_tmp->list));

	ifle_tmp->index = ifim->ifi_index;
	ifle_tmp->flags = ifim->ifi_flags;
	ifle_tmp->type = ifim->ifi_type;

	/* parse interface infomation message attributes */
	parse_ifinfo(ifla, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ifimsg(ifim, ifla);
	}

	/* get interface name */
	if(ifla[IFLA_IFNAME]) {
		strncpy(ifle_tmp->name, RTA_DATA(ifla[IFLA_IFNAME]), IFNAMSIZ);
		mp += sprintf(mp, "if=%s ", ifle_tmp->name);
	}

	/* get interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		memcpy(ifle_tmp->addr, addrstr, sizeof(addrstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "local=%s ", ifle_tmp->addr);
				break;
			default:
				mp += sprintf(mp, "lladdr=%s ", ifle_tmp->addr);
		}
	}

	/* get broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		memcpy(ifle_tmp->brd, brdstr, sizeof(brdstr));
		switch(ifle_tmp->type) {
			case ARPHRD_TUNNEL:
			case ARPHRD_IPGRE:
			case ARPHRD_SIT:
			case ARPHRD_TUNNEL6:
				mp += sprintf(mp, "remote=%s ", ifle_tmp->brd);
				break;
		}
	}

	/* get interface MTU */
	if(ifla[IFLA_MTU]) {
		ifle_tmp->mtu = *(unsigned *)RTA_DATA(ifla[IFLA_MTU]);
		mp += sprintf(mp, "mtu=%hu ", ifle_tmp->mtu);
	}

	/* get interface information */
	if(ifla[IFLA_LINKINFO]) {
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		if (linkinfo[IFLA_INFO_KIND]) {
			if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
				ifle_tmp->vid = *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]);
				mp += sprintf(mp, "vlan=%hu ", ifle_tmp->vid);
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "bond")) {
				mp += sprintf(mp, "info=bond ");
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "tun")) {
				mp += sprintf(mp, "info=tun ");
			}
		}
	}

	/* check protocol family(PF_UNSPEC only) */
	if(ifim->ifi_family != PF_UNSPEC)
		return(0);

	/* check nlmsg_type */
	if(nlh->nlmsg_type == RTM_NEWLINK) {
		/* search interface list entry */
		ifle = search_iflist(ifle_tmp->index);
		if(ifle) {
			/* check administrative status */
			if((ifle->flags & IFF_UP) && !(ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to disabled", msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_UP) && (ifle_tmp->flags & IFF_UP))
				rec_log("[%-12s] interface %s state changed to enabled", msg_type, ifle_tmp->name);

			/* check operational status */
			if((ifle->flags & IFF_RUNNING) && !(ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to down", msg_type, ifle_tmp->name);
			else if(!(ifle->flags & IFF_RUNNING) && (ifle_tmp->flags & IFF_RUNNING))
				rec_log("[%-12s] interface %s state changed to up", msg_type, ifle_tmp->name);

			ifle->flags = ifle_tmp->flags;

			/* check interface name */
			if(ifla[IFLA_IFNAME]) {
				if(strncmp(ifle->name, ifle_tmp->name, IFNAMSIZ)) {
					rec_log("[%-12s] interface name changed from %s to %s",
						msg_type, ifle->name, ifle_tmp->name);
					strncpy(ifle->name, ifle_tmp->name, IFNAMSIZ);
				}
			}

			/* check interface address */
			if(ifla[IFLA_ADDRESS]) {
				if(memcmp(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s local address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
							break;
						default:
							rec_log("[%-12s] interface %s link layer address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->addr, ifle_tmp->addr);
					}
					memcpy(ifle->addr, ifle_tmp->addr, sizeof(ifle->addr));
				}
			}

			/* check broadcast address */
			if(ifla[IFLA_BROADCAST]) {
				if(memcmp(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd))) {
					switch(ifle_tmp->type) {
						case ARPHRD_TUNNEL:
						case ARPHRD_IPGRE:
						case ARPHRD_SIT:
						case ARPHRD_TUNNEL6:
							rec_log("[%-12s] interface %s remote address changed "
								"from %s to %s",
								msg_type, ifle->name, ifle->brd, ifle_tmp->brd);
							break;
					}
					memcpy(ifle->brd, ifle_tmp->brd, sizeof(ifle->brd));
				}
			}

			ifle->type = ifle_tmp->type;

			/* check interface MTU */
			if(ifla[IFLA_MTU]) {
				if(ifle->mtu != ifle_tmp->mtu) {
					rec_log("[%-12s] interface %s mtu changed from %hu to %hu",
						msg_type, ifle->name, ifle->mtu, ifle_tmp->mtu);
					ifle->mtu = ifle_tmp->mtu;
				}
			} 

			/* check interface vlan id */
			if(ifle->vid != ifle_tmp->vid) {
				rec_log("[%-12s] interface %s vlan id changed from %hu to %hu",
					msg_type, ifle->name, ifle->vid, ifle_tmp->vid);
				ifle->vid = ifle_tmp->vid;
			}

			free(ifle_tmp);
		} else {
			/* add interface list entry*/
			add_iflist_entry(&(ifle_tmp->list));

			/* check interface state */
			char state[20] = "";

			/* check administrative state */
			(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

			/* check operational state */
			(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

			rec_log("[%-12s] interface added: %sstate=%s", msg_type, msg, state);
		}
	} else if(nlh->nlmsg_type == RTM_DELLINK) {
		/* move entry from interface list to interface history */
		move_iflist_entry(ifle_tmp->index);
		del_ifhist_entry();

		/* delete interface information list */
		//del_iflist_entry(ifle_tmp->index);

		/* check interface state */
		char state[20] = "";

		/* check administrative state */
		(ifle_tmp->flags & IFF_UP) ? strcpy(state, "enabled,") : strcpy(state, "disabled,");

		/* check operational state */
		(ifle_tmp->flags & IFF_RUNNING) ? strcat(state, "linkup") : strcat(state, "linkdown");

		rec_log("[%-12s] interface deleted: %sstate=%s", msg_type, msg, state);

		free(ifle_tmp);
	} else {
		rec_log("[%-12s] unkonwn nlmsg_type: %d", "ERROR", nlh->nlmsg_type);
		free(ifle_tmp);
	}

	return(0);
}

/*
 * debug interface information messages
 */ 
void debug_ifimsg(struct ifinfomsg *ifim, struct rtattr *ifla[])
{
	unsigned flags = ifim->ifi_flags;
	char flist[136] = "";
	int i;
	char *p;

#define _IFF(f) \
	if(flags & IFF_##f) \
		(flags &= ~IFF_##f) ? strcat(flist, #f ",") : strcat(flist, #f);
	_IFF(UP);
	_IFF(BROADCAST);
	_IFF(DEBUG);
	_IFF(LOOPBACK);
	_IFF(POINTOPOINT);
	_IFF(NOTRAILERS);
	_IFF(RUNNING);
	_IFF(NOARP);
	_IFF(PROMISC);
	_IFF(ALLMULTI);
	_IFF(MASTER);
	_IFF(SLAVE);
	_IFF(MULTICAST);
	_IFF(PORTSEL);
	_IFF(AUTOMEDIA);
	_IFF(DYNAMIC);
	_IFF(LOWER_UP);
	_IFF(DORMANT);
	_IFF(ECHO);
#undef _IFF

	/* debug ifinfomsg */
	rec_dbg("******** ifinfomsg **************************************************");
	rec_dbg("ifinfomsg:");
	rec_dbg("    ifi_family: %d(%s)", ifim->ifi_family, af_type_n2s(ifim->ifi_family));
	rec_dbg("    ifi_type: %hu(%s)", ifim->ifi_type, ifi_type_n2s(ifim->ifi_type));

	char ifname[IFNAMSIZ] = "";
	if_indextoname(ifim->ifi_index, ifname);
	rec_dbg("    ifi_index: %d(%s)", ifim->ifi_index, ifname);
	rec_dbg("    ifi_flags: 0x%.8x(%s)", ifim->ifi_flags, flist);
	rec_dbg("    ifi_change: %u", ifim->ifi_change);

	/* debug interface link attributes */
	rec_dbg("******** attributes *************************************************");

	/* debug interface address */
	if(ifla[IFLA_ADDRESS]) {
		char addrstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_ADDRESS]),
			RTA_PAYLOAD(ifla[IFLA_ADDRESS]), addrstr, sizeof(addrstr));
		rec_dbg("IFLA_ADDRESS(%hu): %s", ifla[IFLA_ADDRESS]->rta_len, addrstr);
	}

	/* debug interface broadcast address */
	if(ifla[IFLA_BROADCAST]) {
		char brdstr[INET6_ADDRSTRLEN+1] = "";
		if_addr_n2s(ifim->ifi_type, (unsigned char *)RTA_DATA(ifla[IFLA_BROADCAST]),
			RTA_PAYLOAD(ifla[IFLA_BROADCAST]), brdstr, sizeof(brdstr));
		rec_dbg("IFLA_BROADCAST(%hu): %s", ifla[IFLA_BROADCAST]->rta_len, brdstr);
	}

	/* debug interface name */
	if(ifla[IFLA_IFNAME])
		rec_dbg("IFLA_IFNAME(%hu): %s", ifla[IFLA_IFNAME]->rta_len, (char *)RTA_DATA(ifla[IFLA_IFNAME]));

	/* debug interface MTU */
	if(ifla[IFLA_MTU])
		rec_dbg("IFLA_MTU(%hu): %hu", ifla[IFLA_MTU]->rta_len, *(unsigned short *)RTA_DATA(ifla[IFLA_MTU]));

	/* debug physical interface index */
	if(ifla[IFLA_LINK])
		rec_dbg("IFLA_LINK(%hu): %u", ifla[IFLA_LINK]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_LINK]));

	/* debug interface queue discipline */
	if(ifla[IFLA_QDISC])
		rec_dbg("IFLA_QDISC(%hu): %s", ifla[IFLA_QDISC]->rta_len, (char *)RTA_DATA(ifla[IFLA_QDISC]));

	/* debug interface statistics */
	if(ifla[IFLA_STATS]) {
		struct rtnl_link_stats *stats = (struct rtnl_link_stats *)RTA_DATA(ifla[IFLA_STATS]);
		rec_dbg("IFLA_STATS(%hu):", ifla[IFLA_STATS]->rta_len);
		rec_dbg("    rtnl_link_stats:");
		rec_dbg("        rx_packets: %u", stats->rx_packets);
		rec_dbg("        tx_packets: %u", stats->tx_packets);
		rec_dbg("        rx_bytes: %u", stats->rx_bytes);
		rec_dbg("        tx_bytes: %u", stats->tx_bytes);
		rec_dbg("        rx_errors: %u", stats->rx_errors);
		rec_dbg("        tx_errors: %u", stats->tx_errors);
		rec_dbg("        rx_dropped: %u", stats->rx_dropped);
		rec_dbg("        tx_dropped: %u", stats->tx_dropped);
		rec_dbg("        multicast: %u", stats->multicast);
		rec_dbg("        collisions: %u", stats->collisions);
		rec_dbg("        rx_length_errors: %u", stats->rx_length_errors);
		rec_dbg("        rx_over_errors: %u", stats->rx_over_errors);
		rec_dbg("        rx_crc_errors: %u", stats->rx_crc_errors);
		rec_dbg("        rx_frame_errors: %u", stats->rx_frame_errors);
		rec_dbg("        rx_fifo_errors: %u", stats->rx_fifo_errors);
		rec_dbg("        rx_missed_errors: %u", stats->rx_missed_errors);
		rec_dbg("        tx_aborted_errors: %u", stats->tx_aborted_errors);
		rec_dbg("        tx_carrier_errors: %u", stats->tx_carrier_errors);
		rec_dbg("        tx_fifo_errors: %u", stats->tx_fifo_errors);
		rec_dbg("        tx_heartbeat_errors: %u", stats->tx_heartbeat_errors);
		rec_dbg("        tx_window_errors: %u", stats->tx_window_errors);
		rec_dbg("        rx_compressed: %u", stats->rx_compressed);
		rec_dbg("        tx_compressed: %u", stats->tx_compressed);
	}

	/* unknown attribute */
	if(ifla[IFLA_COST])
		rec_dbg("IFLA_COST(%hu): %u", ifla[IFLA_COST]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_COST]));

	/* unknown attribute */
	if(ifla[IFLA_PRIORITY])
		rec_dbg("IFLA_PRIORITY(%hu): %u",
			ifla[IFLA_PRIORITY]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_PRIORITY]));

	/* debug master(parent) interface index */
	if(ifla[IFLA_MASTER])
		rec_dbg("IFLA_MASTER(%hu): %u", ifla[IFLA_MASTER]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_MASTER]));

	/* debug wireless interface information(struct iw_event) */
	if(ifla[IFLA_WIRELESS])
		rec_dbg("IFLA_WIRELESS(%hu): -- ignored --", ifla[IFLA_WIRELESS]->rta_len);

	/* debug interface protocol information */
	if(ifla[IFLA_PROTINFO])
		rec_dbg("IFLA_PROTINFO(%hu): -- ignored --", ifla[IFLA_PROTINFO]->rta_len);

	/* debug interafce transmit queue length */
	if(ifla[IFLA_TXQLEN])
		rec_dbg("IFLA_TXQLEN(%hu): %u", ifla[IFLA_TXQLEN]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_TXQLEN]));

	/* debug interface memory mapping */
	if(ifla[IFLA_MAP]) {
		struct rtnl_link_ifmap *ifmap = (struct rtnl_link_ifmap *)RTA_DATA(ifla[IFLA_MAP]);
		unsigned char *mem = (unsigned char *)&(ifmap->mem_start);
		rec_dbg("IFLA_MAP(%hu):", ifla[IFLA_MAP]->rta_len);
		rec_dbg("    rtnl_link_ifmap:");
		rec_dbg("        mem_start: 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->mem_end);
		rec_dbg("        mem_end: 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		mem = (unsigned char *)&(ifmap->base_addr);
		rec_dbg("        base_addr: 0x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
			mem[0], mem[1], mem[2], mem[3], mem[4], mem[5], mem[6], mem[7]);
		rec_dbg("        irq: %hu", ifmap->irq);
		rec_dbg("        dma: 0x%.2x", ifmap->dma);
		rec_dbg("        port: 0x%.2x", ifmap->port);
	}

	/* unknown attribute */
	if(ifla[IFLA_WEIGHT])
		rec_dbg("IFLA_WEIGHT(%hu): %u", ifla[IFLA_WEIGHT]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_WEIGHT]));

	/* debug interface operational state(RFC 2863) */
	if(ifla[IFLA_OPERSTATE])
		rec_dbg("IFLA_OPERSTATE(%hu): %d(%s)",
			ifla[IFLA_OPERSTATE]->rta_len, *(int *)RTA_DATA(ifla[IFLA_OPERSTATE]),
			if_oper_state_n2s(*(int *)RTA_DATA(ifla[IFLA_OPERSTATE])));

	/* debug interface link mode */
	if(ifla[IFLA_LINKMODE])
		rec_dbg("IFLA_LINKMODE(%hu): %d(%s)",
			ifla[IFLA_LINKMODE]->rta_len, *(int *)RTA_DATA(ifla[IFLA_LINKMODE]),
			if_link_mode_n2s(*(int *)RTA_DATA(ifla[IFLA_LINKMODE])));

	/*
	 * debug interface link information
	 *
	 * IFLA_LINKINFO was committed in kernel 2.6.23.
	 * commit 38f7b870d4a6a5d3ec21557e849620cb7d032965
	 * [RTNETLINK]: Link creation API
	 */
	if(ifla[IFLA_LINKINFO]) {
		struct rtattr *linkinfo[__IFLA_INFO_MAX];
		struct rtattr *vlaninfo[__IFLA_VLAN_MAX];
		parse_linkinfo(linkinfo, ifla[IFLA_LINKINFO]);
		rec_dbg("IFLA_LINKINFO(%hu):", ifla[IFLA_LINKINFO]->rta_len);
		rec_dbg("    IFLA_INFO_KIND(%hu): %s",
			linkinfo[IFLA_INFO_KIND]->rta_len, (char *)RTA_DATA(linkinfo[IFLA_INFO_KIND]));
		if (linkinfo[IFLA_INFO_KIND]) {
			if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "vlan")) {
				parse_vlaninfo(vlaninfo, linkinfo[IFLA_INFO_DATA]);
				rec_dbg("        IFLA_VLAN_ID(%hu): %d",
					vlaninfo[IFLA_INFO_KIND]->rta_len, *(int *)RTA_DATA(vlaninfo[IFLA_VLAN_ID]));
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "bond")) {
				/* ignore */
			} else if(!strcmp(RTA_DATA(linkinfo[IFLA_INFO_KIND]), "tun")) {
				/* ignore */
			}
		}
	}

	/* 
	 * debug pid of process in network name space
	 *
	 * IFLA_NET_NS_PID was committed in kernel 2.6.24.
	 * commit d8a5ec672768c3cf4d51d7a63fc071520afa1617
	 * [NET]: netlink support for moving devices between network namespaces.
	 */
	if(ifla[IFLA_NET_NS_PID])
		rec_dbg("IFLA_NET_NS_PID(%hu): %u",
			ifla[IFLA_NET_NS_PID]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_NET_NS_PID]));

	/*
	 * debug interface alias name
	 *
	 * IFLA_IFALIAS was committed in kernel 2.6.28.
	 * commit 0b815a1a6d43ab498674b8430c8c35ab08487a16
	 * net: network device name ifalias support
	 */ 
	if(ifla[IFLA_IFALIAS])
		rec_dbg("IFLA_IFALIAS(%hu): %s", ifla[IFLA_IFALIAS]->rta_len, (char *)RTA_DATA(ifla[IFLA_IFALIAS]));

	/*
	 * debug number of VFs
	 *
	 * IFLA_NUM_VF was committed in kernel 2.6.34.
	 * commit b280da8d54b8d82b52f368a8703b7ada6c1744d5
	 * if_link: Add SR-IOV configuration methods
	 */
	if(ifla[IFLA_NUM_VF])
		rec_dbg("IFLA_NUM_VF(%hu): %u", ifla[IFLA_NUM_VF]->rta_len, *(unsigned *)RTA_DATA(ifla[IFLA_NUM_VF]));

	/*
	 * debug VFs information list
	 *
	 * IFLA_VFINFO_LIST was committed in kernel 2.6.34.
	 * commit c02db8c6290bb992442fec1407643c94cc414375
	 * rtnetlink: make SR-IOV VF interface symmetric
	 */ 
	if(ifla[IFLA_VFINFO_LIST])
		rec_dbg("IFLA_VFINFO_LIST(%hu): -- ignored --", ifla[IFLA_VFINFO_LIST]->rta_len);

	rec_dbg("");

	return;
}

/*
 * convert an interafce type from number to string
 */ 
const char *ifi_type_n2s(int type)
{
	struct {
		int type;
		char *name;
	} ifi_tnames[] = {
#define _IFI_T(x) { x, #x }
		_IFI_T(ARPHRD_NETROM),
		_IFI_T(ARPHRD_ETHER),
		_IFI_T(ARPHRD_EETHER),
		_IFI_T(ARPHRD_AX25),
		_IFI_T(ARPHRD_PRONET),
		_IFI_T(ARPHRD_CHAOS),
		_IFI_T(ARPHRD_IEEE802),
		_IFI_T(ARPHRD_ARCNET),
		_IFI_T(ARPHRD_APPLETLK),
		_IFI_T(ARPHRD_DLCI),
		_IFI_T(ARPHRD_ATM),
		_IFI_T(ARPHRD_METRICOM),
		_IFI_T(ARPHRD_IEEE1394),
		_IFI_T(ARPHRD_EUI64),
		_IFI_T(ARPHRD_INFINIBAND),
		_IFI_T(ARPHRD_SLIP),
		_IFI_T(ARPHRD_CSLIP),
		_IFI_T(ARPHRD_SLIP6),
		_IFI_T(ARPHRD_CSLIP6),
		_IFI_T(ARPHRD_RSRVD),
		_IFI_T(ARPHRD_ADAPT),
		_IFI_T(ARPHRD_ROSE),
		_IFI_T(ARPHRD_X25),
		_IFI_T(ARPHRD_HWX25),
		_IFI_T(ARPHRD_CAN),
		_IFI_T(ARPHRD_PPP),
		_IFI_T(ARPHRD_CISCO),
		_IFI_T(ARPHRD_HDLC),
		_IFI_T(ARPHRD_LAPB),
		_IFI_T(ARPHRD_DDCMP),
		_IFI_T(ARPHRD_RAWHDLC),
		_IFI_T(ARPHRD_TUNNEL),
		_IFI_T(ARPHRD_TUNNEL6),
		_IFI_T(ARPHRD_FRAD),
		_IFI_T(ARPHRD_SKIP),
		_IFI_T(ARPHRD_LOOPBACK),
		_IFI_T(ARPHRD_LOCALTLK),
		_IFI_T(ARPHRD_FDDI),
		_IFI_T(ARPHRD_BIF),
		_IFI_T(ARPHRD_SIT),
		_IFI_T(ARPHRD_IPDDP),
		_IFI_T(ARPHRD_IPGRE),
		_IFI_T(ARPHRD_PIMREG),
		_IFI_T(ARPHRD_HIPPI),
		_IFI_T(ARPHRD_ASH),
		_IFI_T(ARPHRD_ECONET),
		_IFI_T(ARPHRD_IRDA),
		_IFI_T(ARPHRD_FCPP),
		_IFI_T(ARPHRD_FCAL),
		_IFI_T(ARPHRD_FCPL),
		_IFI_T(ARPHRD_FCFABRIC),
		_IFI_T(ARPHRD_IEEE802_TR),
		_IFI_T(ARPHRD_IEEE80211),
		_IFI_T(ARPHRD_IEEE80211_PRISM),
		_IFI_T(ARPHRD_IEEE80211_RADIOTAP),
		_IFI_T(ARPHRD_IEEE802154),
		_IFI_T(ARPHRD_PHONET),
		_IFI_T(ARPHRD_PHONET_PIPE)
#undef _IFI_T
	};
	int i;

	for(i = 0; i < sizeof(ifi_tnames)/sizeof(ifi_tnames[0]); i++)
		if(ifi_tnames[i].type == type)
			return ifi_tnames[i].name;

	return "unknown type";
}

/*
 * convert an interafce operational state from number to string
 */
const char *if_oper_state_n2s(int state)
{
	struct {
		int state;
		char *name;
	} if_osnames[] = {
#define _IF_OS(x) { x, #x }
		_IF_OS(IF_OPER_UNKNOWN),
		_IF_OS(IF_OPER_NOTPRESENT),
		_IF_OS(IF_OPER_DOWN),
		_IF_OS(IF_OPER_LOWERLAYERDOWN),
		_IF_OS(IF_OPER_TESTING),
		_IF_OS(IF_OPER_DORMANT),
		_IF_OS(IF_OPER_UP)
#undef _IF_OS
	};
	int i;

	for(i = 0; i < sizeof(if_osnames)/sizeof(if_osnames[0]); i++)
		if(if_osnames[i].state == state)
			return if_osnames[i].name;

	return "unknown state";
}

/*
 * convert an interafce link mode from number to string
 */
const char *if_link_mode_n2s(int mode)
{
	struct {
		int mode;
		char *name;
	} if_lmnames[] = {
#define _IF_LM(x) { x, #x }
		_IF_LM(IF_LINK_MODE_DEFAULT),
		_IF_LM(IF_LINK_MODE_DORMANT)
#undef _IF_LM
	};
	int i;

	for(i = 0; i < sizeof(if_lmnames)/sizeof(if_lmnames[0]); i++)
		if(if_lmnames[i].mode == mode)
			return if_lmnames[i].name;

	return "unknown mode";
}

/*
 * convert an interface address from number to string
 */ 
void if_addr_n2s(unsigned short type, unsigned char *addr, int addrlen, char *addrstr, int addrstrlen)
{
	int i;

	switch(type) {
		case ARPHRD_TUNNEL:
		case ARPHRD_IPGRE:
		case ARPHRD_SIT:
			inet_ntop(AF_INET, addr, addrstr, addrstrlen);
			break;
		case ARPHRD_TUNNEL6:
			inet_ntop(AF_INET6, addr, addrstr, addrstrlen);
			break;
		default:
			for(i = 0; i < addrlen; i++)
				addrstr += sprintf(addrstr, "%.2x:", addr[i]);
			addrstr--;
			*addrstr = '\0';
	}
}
