#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/* top of interface list */
static LIST_HEAD(head);

/*
 * create a neighbour discovery cache list
 */ 
int create_ndlist(struct msghdr *msg)
{
	struct nlmsghdr *nlh;
	int nlh_len;
	struct ndmsg *ndm;
	struct rtattr *nda[__NDA_MAX];
	struct ndlist_entry *ndle;

	/* get netlink message header */
	nlh = msg->msg_iov->iov_base;
	nlh_len = msg->msg_iov->iov_len;

	/* parse netlink message header */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		/* whether netlink message header ends or not */
		if(nlh->nlmsg_type == NLMSG_DONE)
			return(1);
	
		/* get neighbour discovery message */
		ndm = (struct ndmsg *)NLMSG_DATA(nlh);

		/* check address family */
		if((ndm->ndm_family != AF_INET) && (ndm->ndm_family != AF_INET6))
			continue;

		/* check address type */
		if(ndm->ndm_type != RTN_UNICAST)
			continue;

		/* parse neighbour discovery attributes */
		parse_ndisc(nda, nlh);

		/* create neighbour discovery list */
		ndle = (struct ndlist_entry *)malloc(sizeof(struct ndlist_entry));
		if(ndle == NULL) {
			rec_log("[%-12s] malloc: failed", "ERROR");
			return(-1);
		}
		memset(ndle, 0, sizeof(struct ndlist_entry));
		INIT_LIST_HEAD(&(ndle->list));

		/* get interface index */
		ndle->ifindex = ndm->ndm_ifindex;

		/* convert interface index to name */
		if_indextoname(ndm->ndm_ifindex, ndle->ifname);

		/* get ip address */
		if(nda[NDA_DST])
			inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]),
				ndle->ipaddr, sizeof(ndle->ipaddr));
	
		/* get link local address */
		if(nda[NDA_LLADDR])
			memcpy(ndle->lladdr, RTA_DATA(nda[NDA_LLADDR]), sizeof(ndle->lladdr));

		/* debug */
		if(optflag & F_DEBUG) {
			debug_nlmsg(nlh);
			debug_ndmsg(ndm, nda);
		}

		/* add neighbour discovery cache list */
		list_add(&(ndle->list), &head);
	}

	//print_ndlist();

	return(0);
}

/*
 * print a neighbour discovery cache list
 */ 
void print_ndlist(void)
{
	struct list_head *l;

	list_for_each(l, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);

		printf("ifindex: %d, ", e->ifindex);
		printf("ifname: %s, ", e->ifname);
		printf("ipaddr: %s, ", e->ipaddr);
		printf("lladdr: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n",
			e->lladdr[0], e->lladdr[1], e->lladdr[2], e->lladdr[3], e->lladdr[4], e->lladdr[5]);
	}
}

/*
 * search an entry in a neighbour discovery cache list
 */ 
static inline struct ndlist_entry *search_ndlist(char *ipa)
{
	struct list_head *l;

	list_for_each(l, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);
		if(!memcmp(ipa, &(e->ipaddr), sizeof(e->ipaddr)))
			return e;
	}

	return(NULL);
}

/*
 * add a neighbour discovery cache to a list
 */
static inline void add_ndlist_entry(struct list_head *l)
{
	INIT_LIST_HEAD(l);
	list_add(l, &head);
}

/*
 * [not in use]
 * delete an entry in a interface list entry
 */
static inline void del_iflist_entry(char *ipa)
{
	struct list_head *l, *n;

	list_for_each_safe(l, n, &head) {
		struct ndlist_entry *e;

		e = list_entry(l, struct ndlist_entry, list);
		if(!memcmp(ipa, &(e->ipaddr), sizeof(e->ipaddr))) {
			list_del(l);
			free(e);
		}
	}
}

/*
 * parse neighbour discovery messages
 */
int parse_ndmsg(struct nlmsghdr *nlh)
{
	struct ndmsg *ndm = (struct ndmsg *)NLMSG_DATA(nlh);
	struct rtattr *nda[__NDA_MAX];
	struct ndlist_entry *ndle_tmp, *ndle;
	char msg_type[16] = "";
	char ndm_type[16] = "";

	/* check address family */
	if(ndm->ndm_family == AF_INET) {
		if((optflag & F_IPV4) || !(optflag & F_IPV6)) {
			strcpy(msg_type, "IPv4 ARP");
			strcpy(ndm_type, "arp cache");
		} else {
			return(1);
		}
	} else if(ndm->ndm_family == AF_INET6) {
		if((optflag & F_IPV6) || !(optflag & F_IPV4)) {
			strcpy(msg_type, "IPv6 NDP");
			strcpy(ndm_type, "ndp cache");
		} else {
			return(1);
		}
	} else {
		rec_log("[%-12s] unknown address family: %d", "ERROR", ndm->ndm_family);
		return(1);
	}

	/* unicast only */
	if(ndm->ndm_type != RTN_UNICAST)
		return(1);

	/* create temporary neighbour discovery entry */
	ndle_tmp = malloc(sizeof(struct ndlist_entry));
	if(!ndle_tmp) {
		rec_log("[%-12s] malloc() failed", "ERROR");
		return(1);
	}
	memset(ndle_tmp, 0, sizeof(struct ndlist_entry));

	/* get interface index */
	ndle_tmp->ifindex = ndm->ndm_ifindex;

	/* convert interface index to name */
	if_indextoname((unsigned)ndm->ndm_ifindex, ndle_tmp->ifname);

	/* parse neighbour discovery attributes */
	parse_ndisc(nda, nlh);

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_ndmsg(ndm, nda);
	}

	/* get ip address */
	if(nda[NDA_DST])
		inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]),
			ndle_tmp->ipaddr, sizeof(ndle_tmp->ipaddr));

	/* get link local address */
	if(nda[NDA_LLADDR])
		memcpy(ndle_tmp->lladdr, RTA_DATA(nda[NDA_LLADDR]), sizeof(ndle_tmp->lladdr));

	/* logging neighbor discovery message */
	if(nlh->nlmsg_type == RTM_NEWNEIGH) {
		/* search neighbour discovery entry by ip address */
		ndle = search_ndlist(ndle_tmp->ipaddr);
		switch(ndm->ndm_state) {
			case NUD_FAILED:
				/* change or add neighbour discovery list entry */
				if(ndle) {
					if(memcmp(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr))) {
						memcpy(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr));
						rec_log("[%-12s] %s invalidated: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, ndle->ifname, ndle->ipaddr,
							ndle->lladdr[0], ndle->lladdr[1], ndle->lladdr[2],
							ndle->lladdr[3], ndle->lladdr[4], ndle->lladdr[5]);
					}
					free(ndle_tmp);
				} else {
					add_ndlist_entry(&(ndle_tmp->list));
					rec_log("[%-12s] %s invalidated: if=%s ipaddr=%s "
						"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
						msg_type, ndm_type, ndle_tmp->ifname, ndle_tmp->ipaddr,
						ndle_tmp->lladdr[0], ndle_tmp->lladdr[1], ndle_tmp->lladdr[2],
						ndle_tmp->lladdr[3], ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
				}
				break;
			case NUD_PERMANENT:
			case NUD_PROBE:
			case NUD_DELAY:
			case NUD_STALE:
			case NUD_REACHABLE:
				/* change or add neighbour discovery list entry */
				if(ndle) {
					if(memcmp(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr))) {
						memcpy(ndle->lladdr, ndle_tmp->lladdr, sizeof(ndle->lladdr));
						rec_log("[%-12s] %s changed: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, ndle->ifname, ndle->ipaddr,
							ndle->lladdr[0], ndle->lladdr[1], ndle->lladdr[2],
							ndle->lladdr[3], ndle->lladdr[4], ndle->lladdr[5]);
					}
					free(ndle_tmp);
				} else {
					add_ndlist_entry(&(ndle_tmp->list));
					rec_log("[%-12s] %s added: if=%s ipaddr=%s "
						"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
						msg_type, ndm_type, ndle_tmp->ifname, ndle_tmp->ipaddr,
						ndle_tmp->lladdr[0], ndle_tmp->lladdr[1], ndle_tmp->lladdr[2],
						ndle_tmp->lladdr[3], ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
				}
				break;
			case NUD_NOARP:
			case NUD_INCOMPLETE:
				break;
		}
	} else if(nlh->nlmsg_type == RTM_DELNEIGH) {
		struct list_head *l, *n;
		struct ndlist_entry *e;

		switch(ndm->ndm_state) {
			case NUD_FAILED:
			case NUD_INCOMPLETE:
			case NUD_PERMANENT:
			case NUD_REACHABLE:
			case NUD_STALE:
			case NUD_DELAY:
			case NUD_PROBE:
			case NUD_NOARP:
				/* search & delete neighbour discovery entry */
				list_for_each_safe(l, n, &head) {
					e = list_entry(l, struct ndlist_entry, list);
					if(!memcmp(e->ipaddr, ndle_tmp->ipaddr, sizeof(e->ipaddr))) {
						rec_log("[%-12s] %s deleted: if=%s ipaddr=%s "
							"lladdr=%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",
							msg_type, ndm_type, e->ifname, ndle_tmp->ipaddr,
							ndle_tmp->lladdr[0], ndle_tmp->lladdr[1],
							ndle_tmp->lladdr[2], ndle_tmp->lladdr[3],
							ndle_tmp->lladdr[4], ndle_tmp->lladdr[5]);
						list_del(l);
						free(e);
						/* because entry matched is only one */
						break;
					}
				}
			break;
		}
		free(ndle_tmp);
	}

	return(0);
}

/*
 * debug neighbour discovery messages
 */
void debug_ndmsg(struct ndmsg *ndm, struct rtattr *nda[])
{
	char str[32];
	char ifname[IFNAMSIZ] = "";
	if_indextoname((unsigned)ndm->ndm_ifindex, ifname);

	/* debug ndmsg */
	rec_dbg("******** ndmsg ******************************************************");
	rec_dbg("ndmsg:");
	rec_dbg("    ndm_family: %d(%s)", ndm->ndm_family, af_type_n2s(ndm->ndm_family));
	rec_dbg("    ndm_ifindex: %d(%s)", ndm->ndm_ifindex, ifname);

	strcpy(str, "");
#define _NUD(s) \
	if(ndm->ndm_state == NUD_##s) \
		strcpy(str, #s);
	_NUD(INCOMPLETE);
	_NUD(REACHABLE);
	_NUD(STALE);
	_NUD(DELAY);
	_NUD(PROBE);
	_NUD(FAILED);
	_NUD(NOARP);
	_NUD(PERMANENT);
#undef _NUD
	rec_dbg("    ndm_state: %d(%s)", ndm->ndm_state, str);

	unsigned char flags = ndm->ndm_flags;
	char flist[36] = "";
#define _NTF(f) \
	if(flags & NTF_##f) \
		(flags &= ~NTF_##f) ? strcat(flist, #f ",") : strcat(flist, #f);
	_NTF(USE);
	_NTF(PROXY);
	_NTF(ROUTER);
#undef _NTF
	rec_dbg("    ndm_flags: %d(%s)", ndm->ndm_flags, flist);

	strcpy(str, "");
#define _RTN(t) \
	if(ndm->ndm_type == RTN_##t) \
		strcpy(str, #t);
	_RTN(UNICAST);
	_RTN(LOCAL);
	_RTN(BROADCAST);
	_RTN(ANYCAST);
	_RTN(MULTICAST);
	_RTN(BLACKHOLE);
	_RTN(UNREACHABLE);
	_RTN(PROHIBIT);
	_RTN(THROW);
	_RTN(NAT);
	_RTN(XRESOLVE);
#undef _RTN
	rec_dbg("    ndm_type: %d(%s)", ndm->ndm_type, str);

	/* debug neighbour discovery attributes */
	rec_dbg("******** attributes *************************************************");

	/* debug ip address */
	if(nda[NDA_DST]) {
		char ipa[INET6_ADDRSTRLEN] = "";
		inet_ntop(ndm->ndm_family, (struct in_addr *)RTA_DATA(nda[NDA_DST]), ipa, sizeof(ipa));
		rec_dbg("NDA_DST(%hu): %s", nda[NDA_DST]->rta_len, ipa);
	}

	/* debug link local address */
	if(nda[NDA_LLADDR]) {
		unsigned char lla[ETHER_ADDR_LEN] = "";
		memcpy(lla, RTA_DATA(nda[NDA_LLADDR]), ETHER_ADDR_LEN);
		rec_dbg("NDA_LLADDR(%hu): %.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
			nda[NDA_LLADDR]->rta_len, lla[0], lla[1], lla[2], lla[3], lla[4], lla[5]);
	}

	/* debug neighbour discovery cache information */
	if(nda[NDA_CACHEINFO]) {
		struct nda_cacheinfo *ndac;
		ndac = (struct nda_cacheinfo *)RTA_DATA(nda[NDA_CACHEINFO]);
		rec_dbg("NDA_CACHEINFO(%hu):", nda[NDA_CACHEINFO]->rta_len);
		rec_dbg("    nda_cacheinfo:");
		rec_dbg("        ndm_confirmed: %u", ndac->ndm_confirmed);
		rec_dbg("        ndm_used: %u", ndac->ndm_used);
		rec_dbg("        ndm_updated: %u", ndac->ndm_updated);
		rec_dbg("        ndm_refcnt: %u", ndac->ndm_refcnt);
	}

	/* debug neighbour discovery probes */
	if(nda[NDA_PROBES])
		rec_dbg("NDA_PROBES(%hu): %u", nda[NDA_PROBES]->rta_len, *(unsigned *)RTA_DATA(nda[NDA_PROBES]));

	rec_dbg("");

	return;
}
