#include "nield.h"
#include "rtnetlink.h"

int optflag;
static int lock_fd;
static char lock_file[MAX_STR_SIZE] = LOCK_FILE;
static char log_file[MAX_STR_SIZE];
static char dbg_file[MAX_STR_SIZE];
volatile int sigterm_received = 0;
volatile int sigint_received = 0;

/*
 * main function
 */
int main(int argc, char *argv[])
{
	int sock = -1, err;
	unsigned groups; 

	/* set option */
	err = set_options(argc, argv);
	if(err < 0)
		close_exit(sock, 0);

	/* open log file */
	err = open_log(log_file);
	if(err < 0)
		close_exit(sock, 0);

	/* open debug file */
	if(optflag & F_DEBUG) {
		err = open_dbg(dbg_file);
		if(err < 0)
			close_exit(sock, 0);
	}

	/* create lock file */
	err = open_lock();
	if(err < 0)
		close_exit(sock, 0);

	/* set signal handlers */
	err = set_signal_handlers();
	if(err < 0)
		close_exit(sock, 0);

	/* initizlize daemon */
	err = init_daemon();
	if(err < 0)
		close_exit(sock, 0);
	
	rec_log("[%-12s] nield v%s started(PID: %ld)", "INFORMATION", NIELD_VERSION, (long)getpid());

	/* write pid to lock file */
	write_lock();

	/* open netlink socket to create list */
	groups = 0;
	sock = open_netlink_socket(groups);
	if(sock < 0)
		close_exit(sock, 1);

	/* request interface list */
	err = send_request(sock, RTM_GETLINK, AF_UNSPEC);
	if(err < 0)
		close_exit(sock, 1);

	/* receive interface list */
	err = recv_reply(sock, RTM_GETLINK);
	if(err != 0)
		close_exit(sock, 1);

	/* request neighbour list */
	err = send_request(sock, RTM_GETNEIGH, AF_UNSPEC);
	if(err < 0)
		close_exit(sock, 1);

	/* receive & create interface list */
	err = recv_reply(sock, RTM_GETNEIGH);
	if(err != 0)
		close_exit(sock, 1);

	/* close socket */
	close(sock);

	/* set rtnetlink multicast groups */
	groups = set_rtnetlink_groups();
	sock = open_netlink_socket(groups);
	if(sock < 0)
		close_exit(sock, 1);

	/* recevie events */
	recv_events(sock);

	close_exit(sock, 1);
}

/*
 * close a rtnetlink socket and a debug file, and then exit
 */
void close_exit(int sock, int log_flag)
{
	/* close file descriptor */
	if(sock > 0)
		close(sock);

	/* close debug file */
	if(optflag & F_DEBUG)
		close_dbg();
	
	if(log_flag)
		rec_log("[%-12s] nield v%s terminated(PID: %ld)", "INFORMATION", NIELD_VERSION, getpid());

	/* close log file */
	close_log();

	/* close lock file */
	close_lock();

	exit(1);
}

/*
 * set options
 */
int set_options(int argc, char *argv[])
{
	int opt;

	/* set default value */
	strcpy(log_file, LOG_FILE_DEFAULT);
	strcpy(dbg_file, DEBUG_FILE_DEFAULT);

	/* parse options */
	while((opt = getopt(argc, argv, "vhl:d:46inarf")) != EOF) {
		switch(opt) {
			case 'v':
				fprintf(stderr, "version: %s\n", NIELD_VERSION);
				return(-1);
			case 'h':
				fprintf(stderr, "usage: nield %s\n", NIELD_USAGE);
				return(-1);
			case 'l':
				if(strlen(optarg) < MAX_STR_SIZE) {
					strcpy(log_file, optarg);
				} else {
					fprintf(stderr, "[ERROR] log file path is longer than %d byte\n", MAX_STR_SIZE);
					return(-1);
				}
				break;
			case 'd':
				if(strlen(optarg) < MAX_STR_SIZE) {
					strcpy(dbg_file, optarg);
				} else {
					fprintf(stderr, "[ERROR] debug file path is longer than %d byte\n", MAX_STR_SIZE);
					return(-1);
				}
				optflag |= F_DEBUG;
				break;
			case '4':
				optflag |= F_IPV4;
				break;
			case '6':
				optflag |= F_IPV6;
				break;
			case 'i':
				optflag |= F_LINK;
				break;
			case 'n':
				optflag |= F_NEIGH;
				break;
			case 'a':
				optflag |= F_IFADDR;
				break;
			case 'r':
				optflag |= F_ROUTE;
				break;
			case 'f':
				optflag |= F_RULE;
				break;
			default:
				return(-1);
		}
	}

	return(0);
}

/*
 * create a lock file
 */
int open_lock(void)
{
	int err;
	char pid_str[16];

	if((lock_fd = open(lock_file, O_RDONLY, 0)) > 0) {
		err = read(lock_fd, pid_str, sizeof(pid_str)-1);
		if(err < 0) {
			fprintf(stderr, "[ERROR] read(): %s\n", strerror(errno));
			fprintf(stderr, "[ERROR] can't read pid file\n");
			close(lock_fd);
			lock_fd = 0;
			return(-1);
		}
		pid_str[err] = '\0';

		if(kill((pid_t)atol(pid_str), 0) == 0) {
			fprintf(stderr, "[ERROR] nield already running(PID: %ld)\n", atol(pid_str));
			close(lock_fd);
			lock_fd = 0;
			return(-1);
		}
		close(lock_fd);

		lock_fd = open(lock_file, O_CREAT|O_TRUNC|O_WRONLY, 0644);
	} else {
		lock_fd = open(lock_file, O_CREAT|O_EXCL|O_WRONLY, 0644);
	}

	if(lock_fd < 0) {
		fprintf(stderr, "[ERROR] open(): %s\n", strerror(errno));
		fprintf(stderr, "[ERROR] can't create lock file\n");
		return(-1);
	}

	return(0);
}

/*
 * write PID to a lock file
 */
void write_lock(void)
{
	char pid_str[16];

	snprintf(pid_str, sizeof(pid_str), "%ld\n", (long)getpid());
	write(lock_fd, pid_str, strlen(pid_str));
}

void close_lock(void)
{
	if(lock_fd) {
		close(lock_fd);
		unlink(lock_file);
	}
}

/*
 * initialize daemon
 */
int init_daemon(void)
{
	/* check user */
	if(getuid() != 0) {
		fprintf(stderr, "[ERROR] only root can run nield\n");
		return(-1);
	}

	/* change root directory */
	if(chdir("/") == EOF) {
		fprintf(stderr, "[ERROR] can't change root directory\n");
		fprintf(stderr, "[ERROR] chdir(): %s\n", strerror(errno));
		return(-1);
	}

	/* detach from controlling terminal */
	if(daemon(0, 0) < 0) {
		fprintf(stderr, "[ERROR] can't detach terminal\n");
		fprintf(stderr, "[ERROR] daemon(): %s\n", strerror(errno));
		return(-1);
	}

	return(0);
}

/*
 * set signal handlers(SIGTERM, SIGINT)
 */
int set_signal_handlers(void)
{
	struct sigaction sigterm_act, sigint_act;

	/* set SIGTERM handler */
	sigterm_act.sa_handler = sigterm_handler;
	if(sigfillset(&sigterm_act.sa_mask) < 0) {
		fprintf(stderr, "[ERROR] sigfillset(): %s\n", strerror(errno));
		return(-1);
	}
	sigterm_act.sa_flags = 0;

	if(sigaction(SIGTERM, &sigterm_act, 0) < 0) {
		fprintf(stderr, "[ERROR] sigaction(): %s\n", strerror(errno));
		return(-1);
	}

	/* set SIGINT handler */
	sigint_act.sa_handler = sigint_handler;
	if(sigfillset(&sigint_act.sa_mask) < 0) {
		fprintf(stderr, "[ERROR] sigfillset(): %s\n", strerror(errno));
		return(-1);
	}
	sigint_act.sa_flags = 0;

	if(sigaction(SIGTERM, &sigint_act, 0) < 0) {
		fprintf(stderr, "[ERROR] sigaction(): %s\n", strerror(errno));
		return(-1);
	}

	return(0);
}
		
/*
 * signal handler for SIGTERM
 */
void sigterm_handler(int sig)
{
	rec_log("[%-12s] SIGTERM called(PID: %ld)", "INFORMATION", (long)getpid());
	sigterm_received = 1;
}

/*
 * signal handler for SIGINT
 */
void sigint_handler(int sig)
{
	rec_log("[%-12s] SIGINT called(PID: %ld)", "INFORMATION", (long)getpid());
	sigint_received = 1;
}

int set_rtnetlink_groups(void)
{
	int groups;

	/* set netlink groups flag to receive events */
	groups = RTMGRP_NOTIFY;
	if(optflag & F_LINK)
		groups |= RTMGRP_LINK;
	if(optflag & F_NEIGH)
		groups |= RTMGRP_NEIGH;
	if(optflag & F_IFADDR) {
		if(optflag & F_IPV4)
			groups |= RTMGRP_IPV4_IFADDR;
		if(optflag & F_IPV6)
			groups |= RTMGRP_IPV6_IFADDR;
		if(!(optflag & (F_IPV4 | F_IPV6)))
			groups |= RTMGRP_IPV4_IFADDR | RTMGRP_IPV6_IFADDR;
	}
	if(optflag & F_ROUTE) {
		if(optflag & F_IPV4)
			groups |= RTMGRP_IPV4_ROUTE;
		if(optflag & F_IPV6)
			groups |= RTMGRP_IPV6_ROUTE;
		if(!(optflag & (F_IPV4 | F_IPV6)))
			groups |= RTMGRP_IPV4_ROUTE | RTMGRP_IPV6_ROUTE;
	}
	if(optflag & F_RULE) {
		if(optflag & F_IPV4)
			groups |= RTMGRP_IPV4_RULE;
		if(optflag & F_IPV6)
			groups |= (1 << (RTNLGRP_IPV6_RULE - 1));
		if(!(optflag & (F_IPV4 | F_IPV6)))
			groups |= RTMGRP_IPV4_ROUTE | (1 << (RTNLGRP_IPV6_RULE - 1));
	}
	if(!(optflag & (F_LINK | F_NEIGH | F_IFADDR | F_ROUTE | F_RULE))) {
		if(optflag & F_IPV4)
			groups |= RTMGRP_LINK | RTMGRP_NEIGH |
			          RTMGRP_IPV4_IFADDR | RTMGRP_IPV4_ROUTE | RTMGRP_IPV4_RULE;
		if(optflag & F_IPV6)
			groups |= RTMGRP_LINK | RTMGRP_NEIGH |
			          RTMGRP_IPV6_IFADDR | RTMGRP_IPV6_ROUTE | (1 << (RTNLGRP_IPV6_RULE - 1));
		/* receive all events if the flag is set to only F_DEBUG or none */
		if(!(optflag & (F_IPV4 | F_IPV6)))
			groups |= RTMGRP_LINK | RTMGRP_NEIGH | 
			          RTMGRP_IPV4_IFADDR | RTMGRP_IPV4_ROUTE | RTMGRP_IPV4_RULE |
				  RTMGRP_IPV6_IFADDR | RTMGRP_IPV6_ROUTE | (1 << (RTNLGRP_IPV6_RULE - 1));
	}

	return(groups);
}

/*
 * open a rtnetlink socket
 */
int open_netlink_socket(unsigned groups)
{
	int sock;
	struct sockaddr_nl nlla;

	/* open netlink socket */
	sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if(sock < 0) {
		rec_log("socket(): %s", strerror(errno));
		exit(1);
	}

	/* initialization */
	memset(&nlla, 0, sizeof(struct sockaddr_nl));

	/* set & bind local address */
	nlla.nl_family = AF_NETLINK;
	nlla.nl_pid = getpid();
	nlla.nl_groups = groups;
	bind(sock, (struct sockaddr *)&nlla, sizeof(struct sockaddr_nl));

	return(sock);
}

/*
 * send a request to kernel through rtnetlink socket
 */
int send_request(int sock, int type, int family)
{
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_nl nlla;
	struct {
		struct nlmsghdr nlm;
		struct rtgenmsg rtgm;
	} req;
	int err;

	/* initialization */
	memset(&nlla, 0, sizeof(struct sockaddr_nl));
	memset(&req, 0, sizeof(req));
	memset(&iov, 0, sizeof(struct iovec));
	memset(&msg, 0, sizeof(struct msghdr));

	/* set remote address */
	nlla.nl_family = AF_NETLINK;

	/* set netlink message header */
	req.nlm.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtgenmsg)); /* it's suspicious */
	req.nlm.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
	req.nlm.nlmsg_type = type;

	/* set routing message header */
	req.rtgm.rtgen_family = family;

	/* set message header */
	iov.iov_base = (void *)&req;
	iov.iov_len = sizeof(req);

	msg.msg_name = (void *)&nlla;
	msg.msg_namelen = sizeof(struct sockaddr_nl);
	msg.msg_iov = (void *)&iov;
	msg.msg_iovlen = 1;

	/* send request message */
	err = sendmsg(sock, &msg, 0);
	if(err < 0) {
		rec_log("[%-12s] sendmsg(): %s", "ERROR", strerror(errno));
		return(1);
	}

	return(0);
}

/*
 * receive a reply from kernel through rtnetlink socket
 */
int recv_reply(int sock, int type)
{
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_nl nlla;
	int err, len;
	char buf[16384];

	/* initialization */
	memset(&nlla, 0, sizeof(nlla));
	memset(&iov, 0, sizeof(iov));
	memset(buf, 0, sizeof(buf));
	memset(&msg, 0, sizeof(msg));

	/* set message header */
	msg.msg_name = (void *)&nlla;
	msg.msg_namelen = sizeof(struct sockaddr_nl);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	/* receive loop */
	while(1) {
		memset(buf, 0, sizeof(buf));
		len = recvmsg(sock, &msg, 0);
		if(len < 0) {
			if(errno == EINTR || errno == EAGAIN)
				continue;
			rec_log("[%-12s] recvmsg(): %s", "ERROR", strerror(errno));
			return(1);
		} else if(len == 0) {
			rec_log("[%-12s] recvmsg(): receive EOF\n", "ERROR");
			return(1);
		}

		if(type == RTM_GETLINK) {
			/* create interface list */
			err = create_iflist(&msg);
			if(err > 0)
				break;
			else if(err < 0)
				return(1);
		} else if(type == RTM_GETNEIGH) {
			/* create neighbour discovery list */
			err = create_ndlist(&msg);
			if(err)
				break;
			else if(err < 0)
				return(1);
		}
	}

	return(0);
}

/*
 * receive notifications from kernel through rtnetlink socket
 */
void recv_events(int sock)
{
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_nl nladdr;
	int err, len;
	char buf[8192];

	/* initialization */
	memset(&msg, 0, sizeof(struct msghdr));
	memset(&iov, 0, sizeof(struct iovec));
	memset(&nladdr, 0, sizeof(struct sockaddr_nl));
	memset(buf, 0, sizeof(buf));

	/* set message header */
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);

	msg.msg_name = (void *)&nladdr;
	msg.msg_namelen = sizeof(struct sockaddr_nl);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	/* loop of receive event */
	while(1) {
		/* detect signal received */
		if(sigterm_received || sigint_received)
			break;

		/* clear buffer */
		memset(buf, 0, sizeof(buf));

		/* receive event */
		len = recvmsg(sock, &msg, 0);
		if(len < 0) {
			if(errno == EINTR || errno == EAGAIN)
				continue;

			rec_log("[%-12s] recvmsg(): %s", "ERROR", strerror(errno));
			break;
		} else if(!len) {
			rec_log("[%-12s] recvmsg(): receive EOF", "ERROR", strerror(errno));
			break;
		}

		err = parse_events(&msg);
		if(err < 0)
			break;

	}
}

/*
 * parse messages in received notifications from kernel
 */
int parse_events(struct msghdr *mhdr)
{
	struct nlmsghdr *nlh;
	struct nlmsgerr *nle;
	int nlh_len;

	/* get netlink message header */
	nlh = mhdr->msg_iov->iov_base;
	nlh_len = mhdr->msg_iov->iov_len;

	/* parse netlink message type */
	for( ; NLMSG_OK(nlh, nlh_len); nlh = NLMSG_NEXT(nlh, nlh_len)) {
		switch(nlh->nlmsg_type) {
			/* interface link message */
			case RTM_NEWLINK:
			case RTM_DELLINK:
				parse_ifimsg(nlh);
				break;
			/* interface address message */
			case RTM_NEWADDR:
			case RTM_DELADDR:
				parse_ifamsg(nlh);
				break;
			/* neighbor discovery message */
			case RTM_NEWNEIGH:
			case RTM_DELNEIGH:
				parse_ndmsg(nlh);
				break;
			/* route message */
			case RTM_NEWROUTE:
			case RTM_DELROUTE:
				parse_rtmsg(nlh);
				break;
			/* fib rule header */
			case RTM_NEWRULE:
			case RTM_DELRULE:
				parse_frhdr(nlh);
				break;
			/* error message */
			case NLMSG_ERROR:
				nle = (struct nlmsgerr *)NLMSG_DATA(nlh);
				rec_log("[%-12s] nlmsg error: %s", "ERROR", strerror(nle->error));
				break;
			/* unknown message */
			default:
				rec_log("[%-12s] unkonwn nlsmg_type: %d", "ERROR", (int)nlh->nlmsg_type);
		}
	}

	return(0);
}
