#include "nield.h"
#include "rtnetlink.h"

extern int optflag;

/*
 * parse route messages
 */
int parse_rtmsg(struct nlmsghdr *nlh)
{
	struct rtmsg *rtm = (struct rtmsg *)NLMSG_DATA(nlh);
	struct rtattr *rta[__RTA_MAX];
	char dst[INET6_ADDRSTRLEN] = "";
	char msg_type[16] = "", msg[128] = "";
	char metric[16] = "", type[16] = "", proto[16] = "", table[16] = "";
	char *mp = msg;

	/* check address family */
	if(rtm->rtm_family == AF_INET) {
		strcpy(msg_type, "IPv4 ROUTE");
		strcpy(dst, "0.0.0.0");
	} else if(rtm->rtm_family == AF_INET6) {
		strcpy(msg_type, "IPv6 ROUTE");
		strcpy(dst, "::");
	} else {
		rec_log("[%-12s] unknown address family: %d", "ERROR", rtm->rtm_family);
		return(1);
	}

	/* convert from table id to table name */
	rt_table_n2s(rtm->rtm_table, table);

	/* parse route attributes */
	parse_rtattr(rta, RTA_MAX, RTM_RTA(rtm), RTM_PAYLOAD(nlh));

	/* debug */
	if(optflag & F_DEBUG) {
		debug_nlmsg(nlh);
		debug_rtmsg(rtm, rta);
	}

	/* check route table id(other than RT_TABLE_LOCAL) */
	if(rtm->rtm_table == RT_TABLE_LOCAL)
		return(1);

	/* check route protocol(other than RTPROT_UNSPEC) */
	if(rtm->rtm_protocol == RTPROT_UNSPEC)
		return(1);
	
	/* check route flags(other then RTM_F_CLONED */
	if(rtm->rtm_flags & RTM_F_CLONED)
		return(1);

	/* get destination prefix */
	if(rta[RTA_DST])
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_DST]), dst, sizeof(dst));
	mp += sprintf(mp, "dst=%s/%d ", dst, rtm->rtm_dst_len);

	/* get source prefix */
	if(rta[RTA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_SRC]), src, sizeof(src));
		mp += sprintf(mp, "src=%s/%d ", src, rtm->rtm_src_len);
	}

	/* get preferred source address */
	if(rta[RTA_PREFSRC]) {
		char prefsrc[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_PREFSRC]), prefsrc, sizeof(prefsrc));
		mp += sprintf(mp, "prefsrc=%s ", prefsrc);
	}

	/* get tos */
	if(rtm->rtm_tos)
		mp += sprintf(mp, "tos=0x%.2x ", rtm->rtm_tos);

	/* get ingress interface */
	if(rta[RTA_IIF]) {
		unsigned iifindex;
		char iifname[IFNAMSIZ] = "";

		iifindex = *((unsigned *)RTA_DATA(rta[RTA_IIF]));
		if(nlh->nlmsg_type == RTM_NEWROUTE)
			if_indextoname(iifindex, iifname);
		else if(nlh->nlmsg_type == RTM_DELROUTE)
			if(!if_indextoname_from_iflist(iifindex, iifname))
				if_indextoname_from_ifhist(iifindex, iifname);
		mp += sprintf(mp, "in=%s ", iifname);
	}

	/* get gateway address */
	if(rta[RTA_GATEWAY]) {
		char gw[INET6_ADDRSTRLEN] = "";

		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_GATEWAY]), gw, sizeof(gw));
		mp += sprintf(mp, "gw=%s ", gw);
	}

	/* get egress interface */
	if(rta[RTA_OIF]) {
		unsigned oifindex;
		char oifname[IFNAMSIZ] = "";

		oifindex = *((unsigned *)RTA_DATA(rta[RTA_OIF]));
		if(nlh->nlmsg_type == RTM_NEWROUTE)
			if_indextoname(oifindex, oifname);
		else if(nlh->nlmsg_type == RTM_DELROUTE)
			if(!if_indextoname_from_iflist(oifindex, oifname))
				if_indextoname_from_ifhist(oifindex, oifname);
		mp += sprintf(mp, "out=%s ", oifname);
	}

	/* get priority(but metric) */
	if(rta[RTA_PRIORITY])
		sprintf(metric, "metric=%d ", *((int *)RTA_DATA(rta[RTA_PRIORITY])));

	/* convert route message type */
	rtm_type_n2s(rtm->rtm_type, type);

	/* convert route message protocol */
	rtm_proto_n2s(rtm->rtm_protocol, proto);

	/* get table id & name */
	if(rta[RTA_TABLE])
		rt_table_n2s(*(int *)RTA_DATA(rta[RTA_TABLE]), table);

	/* get multipath */
	if(rta[RTA_MULTIPATH]) {
		struct rtnexthop *rtnh = RTA_DATA(rta[RTA_MULTIPATH]);
		int rtnh_len = RTA_PAYLOAD(rta[RTA_MULTIPATH]);
		struct rtattr *rtna[__RTA_MAX];
		char rtnh_ifname[IFNAMSIZ] = "";
		char rtnh_gw[INET6_ADDRSTRLEN] = "";

		for(; RTNH_OK(rtnh, rtnh_len);
			rtnh = RTNH_NEXT(rtnh), rtnh_len -= RTNH_ALIGN(rtnh->rtnh_len)) {
			parse_rtattr(rtna, RTA_MAX, RTNH_DATA(rtnh), rtnh->rtnh_len - sizeof(*rtnh));
			if(rtna[RTA_GATEWAY])
				inet_ntop(rtm->rtm_family, RTA_DATA(rtna[RTA_GATEWAY]),
					rtnh_gw, sizeof(rtnh_gw));

			/* get interface name & logging routing table message */
			if(nlh->nlmsg_type == RTM_NEWROUTE) {
				if_indextoname(rtnh->rtnh_ifindex, rtnh_ifname);

				rec_log("[%-12s] route added: %sgw=%s out=%s "
					"%sweight=%d type=%s proto=%s table=%s",
					msg_type, msg, rtnh_gw, rtnh_ifname,
					metric, rtnh->rtnh_hops+1, type, proto, table);
			} else if(nlh->nlmsg_type == RTM_DELROUTE) {
				if(!if_indextoname_from_iflist(rtnh->rtnh_ifindex, rtnh_ifname))
					if_indextoname_from_ifhist(rtnh->rtnh_ifindex, rtnh_ifname);

				rec_log("[%-12s] route deleted: %sgw=%s out=%s " 
					"%sweight=%d type=%s proto=%s table=%s",
					msg_type, msg, rtnh_gw, rtnh_ifname,
					metric, rtnh->rtnh_hops+1, type, proto, table);
			}
		}

		return(0);
	}

	/* logging routing message */
	if(nlh->nlmsg_type == RTM_NEWROUTE)
		rec_log("[%-12s] route added: %s%stype=%s proto=%s table=%s",
			msg_type, msg, metric, type, proto, table);
	else if(nlh->nlmsg_type == RTM_DELROUTE)
		rec_log("[%-12s] route deleted: %s%stype=%s proto=%s table=%s",
			msg_type, msg, metric, type, proto, table);

	return(0);
}

/*
 * debug route messages
 */
void debug_rtmsg(struct rtmsg *rtm, struct rtattr *rta[])
{
	char str[16];

	/* debug rtmsg */
	rec_dbg("******** rtmsg ******************************************************");
	rec_dbg("rtmsg:");
	rec_dbg("    rtm_family: %d(%s)", rtm->rtm_family, af_type_n2s(rtm->rtm_family));
	rec_dbg("    rtm_dst_len: %d", rtm->rtm_dst_len);
	rec_dbg("    rtm_src_len: %d", rtm->rtm_src_len);
	rec_dbg("    rtm_tos: %d", rtm->rtm_tos);

	strcpy(str, "");
#define _RT_TABLE(t) \
	if(rtm->rtm_table == RT_TABLE_##t) \
		strcpy(str, #t);
	_RT_TABLE(UNSPEC);
	_RT_TABLE(COMPAT);
	_RT_TABLE(DEFAULT);
	_RT_TABLE(MAIN);
	_RT_TABLE(LOCAL);
#undef _RT_TABLE
	rec_dbg("    rtm_table: %d(%s)", rtm->rtm_table, str);

	strcpy(str, "");
#define _RTPROT(p) \
	if(rtm->rtm_protocol == RTPROT_##p) \
		strcpy(str, #p);
	_RTPROT(UNSPEC);
	_RTPROT(REDIRECT);
	_RTPROT(KERNEL);
	_RTPROT(BOOT);
	_RTPROT(STATIC);
	_RTPROT(GATED);
	_RTPROT(RA);
	_RTPROT(MRT);
	_RTPROT(ZEBRA);
	_RTPROT(BIRD);
	_RTPROT(DNROUTED);
	_RTPROT(XORP);
	_RTPROT(NTK);
	_RTPROT(DHCP);
#undef _RT_PROT
	rec_dbg("    rtm_protocol: %d(%s)", rtm->rtm_protocol, str);

	strcpy(str, "");
#define _RT_SCOPE(s) \
	if(rtm->rtm_scope == RT_SCOPE_##s) \
		strcpy(str, #s);
	_RT_SCOPE(UNIVERSE);
	_RT_SCOPE(SITE);
	_RT_SCOPE(LINK);
	_RT_SCOPE(HOST);
	_RT_SCOPE(NOWHERE);
#undef _RT_SCOPE
	rec_dbg("    rtm_scope: %d(%s)", rtm->rtm_scope, str);

	strcpy(str, "");
#define _RTN_TYPE(t) \
	if(rtm->rtm_type == RTN_##t) \
		strcpy(str, #t);
	_RTN_TYPE(UNSPEC);
	_RTN_TYPE(UNICAST);
	_RTN_TYPE(LOCAL);
	_RTN_TYPE(BROADCAST);
	_RTN_TYPE(ANYCAST);
	_RTN_TYPE(MULTICAST);
	_RTN_TYPE(BLACKHOLE);
	_RTN_TYPE(UNREACHABLE);
	_RTN_TYPE(PROHIBIT);
	_RTN_TYPE(THROW);
	_RTN_TYPE(NAT);
	_RTN_TYPE(XRESOLVE);
#undef _RTN_TYPE
	rec_dbg("    rtm_type: %d(%s)", rtm->rtm_type, str);

	unsigned flags = rtm->rtm_flags;
	char flist[36] = "";

#define _RTM_F(f) \
	if(flags & RTM_F_##f) \
		(flags &= ~RTM_F_##f) ? strcat(flist, #f ", ") : strcat(flist, #f);
	_RTM_F(NOTIFY);
	_RTM_F(CLONED);
	_RTM_F(EQUALIZE);
	_RTM_F(PREFIX);
#undef _RTM_F
	rec_dbg("    rtm_flags: %d(%s)", rtm->rtm_flags, flist);

	/* debug route attributes */
	rec_dbg("******** attributes *************************************************");

	/* debug destination prefix */
	if(rta[RTA_DST]) {
		char dst[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_DST]), dst, sizeof(dst));
		rec_dbg("RTA_DST(%hu): %s", rta[RTA_DST]->rta_len, dst);
	}

	/* debug source prefix */
	if(rta[RTA_SRC]) {
		char src[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_SRC]), src, sizeof(src));
		rec_dbg("RTA_SRC(%hu): %s", rta[RTA_SRC]->rta_len, src);
	}

	/* debug ingress interface */
	if(rta[RTA_IIF]) {
		char iifname[IFNAMSIZ] = "";
		int iifindex = *((int *)RTA_DATA(rta[RTA_IIF]));
		if_indextoname(iifindex, iifname);
		rec_dbg("RTA_IIF(%hu): %d(%s)", rta[RTA_IIF]->rta_len, iifindex, iifname);
	}

	/* debug egress interface */
	if(rta[RTA_OIF]) {
		char oifname[IFNAMSIZ] = "";
		int oifindex = *((int *)RTA_DATA(rta[RTA_OIF]));
		if_indextoname(oifindex, oifname);
		rec_dbg("RTA_OIF(%hu): %d(%s)", rta[RTA_OIF]->rta_len, oifindex, oifname);
	}

	/* debug gateway */
	if(rta[RTA_GATEWAY]) {
		char gw[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_GATEWAY]), gw, sizeof(gw));
		rec_dbg("RTA_GATEWAY(%hu): %s", rta[RTA_GATEWAY]->rta_len, gw);
	}

	/* debug priority(but metric) */
	if(rta[RTA_PRIORITY]) {
		int priority = *((int *)RTA_DATA(rta[RTA_PRIORITY]));
		rec_dbg("RTA_PRIORITY(%hu): %d", rta[RTA_PRIORITY]->rta_len, priority);
	}

	/* debug preferred source address */
	if(rta[RTA_PREFSRC]) {
		char prefsrc[INET6_ADDRSTRLEN] = "";
		inet_ntop(rtm->rtm_family, RTA_DATA(rta[RTA_PREFSRC]), prefsrc, sizeof(prefsrc));
		rec_dbg("RTA_PREFSRC(%hu): %s", rta[RTA_PREFSRC]->rta_len, prefsrc);
	}

	/* debug metric */
	if(rta[RTA_METRICS]) {
		rec_dbg("RTA_METRICS(%hu):", rta[RTA_METRICS]->rta_len);

		struct rtattr *rtax[RTAX_MAX+1];

		parse_rtattr(rtax, RTAX_MAX, RTA_DATA(rta[RTA_METRICS]), RTA_PAYLOAD(rta[RTA_METRICS]));
		if(rtax[RTAX_LOCK])
			rec_dbg("    RTAX_LOCK: %u", *(unsigned *)RTA_DATA(rtax[RTAX_LOCK]));
		if(rtax[RTAX_MTU])
			rec_dbg("    RTAX_MTU: %u", *(unsigned *)RTA_DATA(rtax[RTAX_MTU]));
		if(rtax[RTAX_ADVMSS])
			rec_dbg("    RTAX_ADVMSS: %u", *(unsigned *)RTA_DATA(rtax[RTAX_ADVMSS]));
		if(rtax[RTAX_HOPLIMIT])
			rec_dbg("    RTAX_HOPLIMIT: %d", *(int *)RTA_DATA(rtax[RTAX_HOPLIMIT]));
		if(rtax[RTAX_WINDOW])
			rec_dbg("    RTAX_WINDOW: %u", *(unsigned *)RTA_DATA(rtax[RTAX_WINDOW]));
	}

	/* debug multipath */
	if(rta[RTA_MULTIPATH]) {
		struct rtnexthop *rtnh = RTA_DATA(rta[RTA_MULTIPATH]);
		int rtnh_len = RTA_PAYLOAD(rta[RTA_MULTIPATH]);
		struct rtattr *rtnha[__RTA_MAX];
		char rtnh_ifname[IFNAMSIZ] = "";
		char rtnh_gw[INET6_ADDRSTRLEN] = "";

		rec_dbg("RTA_MULTIPATH(%hu):", rta[RTA_MULTIPATH]->rta_len);

		for(; RTNH_OK(rtnh, rtnh_len);
			rtnh = RTNH_NEXT(rtnh), rtnh_len -= RTNH_ALIGN(rtnh->rtnh_len)) {
			rec_dbg("    rtnexthop:");
			rec_dbg("        rtnh_len: %hu", rtnh->rtnh_len);

			flags = rtnh->rtnh_flags;
			strcpy(flist, "");
#define _RTNH_F(f) \
			if(flags & RTNH_F_##f) \
				(flags &= ~RTNH_F_##f) ? strcat(flist, #f ", ") : strcat(flist, #f);
			_RTNH_F(DEAD);
			_RTNH_F(PERVASIVE);
			_RTNH_F(ONLINK);
#undef _RTNH_F
			rec_dbg("        rtnh_flags: %d(%s)", rtnh->rtnh_flags, flist);
			rec_dbg("        rtnh_hops: %d", rtnh->rtnh_hops);

			if_indextoname(rtnh->rtnh_ifindex, rtnh_ifname);
			rec_dbg("        rtnh_ifindex: %d(%s)", rtnh->rtnh_ifindex, rtnh_ifname);

			parse_rtattr(rtnha, RTA_MAX, RTNH_DATA(rtnh), rtnh->rtnh_len - sizeof(*rtnh));
			if(rtnha[RTA_GATEWAY]) {
				inet_ntop(rtm->rtm_family, RTA_DATA(rtnha[RTA_GATEWAY]),
					rtnh_gw, sizeof(rtnh_gw));
				rec_dbg("    RTA_GATEWAY: %s", rtnh_gw);
			}
		}
	}

	/* debug flow */
	if(rta[RTA_FLOW])
		rec_dbg("RTA_FLOW(%hu): %u", rta[RTA_FLOW]->rta_len, *((unsigned *)RTA_DATA(rta[RTA_FLOW])));

	/* debug route cache information */
	if(rta[RTA_CACHEINFO]) {
		struct rta_cacheinfo *rtac;
		rtac = (struct rta_cacheinfo *)RTA_DATA(rta[RTA_CACHEINFO]);
		rec_dbg("RTA_CACHEINFO(%hu):", rta[RTA_CACHEINFO]->rta_len);
		rec_dbg("    rta_cacheinfo:");
		rec_dbg("        rta_clntref: %u", rtac->rta_clntref);
		rec_dbg("        rta_lastuse: %u", rtac->rta_lastuse);
		rec_dbg("        rta_expires: %d", rtac->rta_expires);
		rec_dbg("        rta_error: %u", rtac->rta_error);
		rec_dbg("        rta_used: %u", rtac->rta_used);
		rec_dbg("        rta_id: %u", rtac->rta_id);
		rec_dbg("        rta_ts: %u", rtac->rta_ts);
		rec_dbg("        rta_tsage: %u", rtac->rta_tsage);
	}

	/* debug route table id & name */
	if(rta[RTA_TABLE]) {
		int table = *((int *)RTA_DATA(rta[RTA_TABLE]));
		strcpy(str, "");
#define _RT_TABLE(t) \
		if(table == RT_TABLE_##t) \
			strcpy(str, #t);
		_RT_TABLE(UNSPEC);
		_RT_TABLE(COMPAT);
		_RT_TABLE(DEFAULT);
		_RT_TABLE(MAIN);
		_RT_TABLE(LOCAL);
#undef _RT_TABLE
		rec_dbg("RTA_TABLE(%hu): %d(%s)", rta[RTA_TABLE]->rta_len, table, str);
	}

	rec_dbg("");

	return;
}

/*
 * convert a route message type from number to string
 */ 
void rtm_type_n2s(unsigned char ntype, char *stype)
{
	switch(ntype) {
		case RTN_UNSPEC:
			strcpy(stype, "none");
			break;
		case RTN_UNICAST:
			strcpy(stype, "unicast");
			break;
		case RTN_LOCAL:
			strcpy(stype, "local");
			break;
		case RTN_BROADCAST:
			strcpy(stype, "broadcast");
			break;
		case RTN_ANYCAST:
			strcpy(stype, "anycast");
			break;
		case RTN_MULTICAST:
			strcpy(stype, "multicast");
			break;
		case RTN_BLACKHOLE:
			strcpy(stype, "blackhole");
			break;
		case RTN_UNREACHABLE:
			strcpy(stype, "unreachable");
			break;
		case RTN_PROHIBIT:
			strcpy(stype, "prohibit");
			break;
		case RTN_THROW:
			strcpy(stype, "throw");
			break;
		case RTN_NAT:
			strcpy(stype, "nat");
			break;
		case RTN_XRESOLVE:
			strcpy(stype, "external");
			break;
		default:
			sprintf(stype, "%d", ntype);
	}
}

/*
 * convert a route message protocol from number to string
 */
void rtm_proto_n2s(int nproto, char *sproto)
{
	switch(nproto) {
		case RTPROT_UNSPEC:
			strcpy(sproto, "none");
			break;
		case RTPROT_REDIRECT:
			strcpy(sproto, "redirect");
			break;
		case RTPROT_KERNEL:
			strcpy(sproto, "kernel");
			break;
		case RTPROT_BOOT:
			strcpy(sproto, "boot");
			break;
		case RTPROT_STATIC:
			strcpy(sproto, "static");
			break;
		case RTPROT_GATED:
			strcpy(sproto, "gated");
			break;
		case RTPROT_RA:
			strcpy(sproto, "ra");
			break;
		case RTPROT_MRT:
			strcpy(sproto, "ra");
			break;
		case RTPROT_ZEBRA:
			strcpy(sproto, "zebra");
			break;
		case RTPROT_BIRD:
			strcpy(sproto, "bird");
			break;
		case RTPROT_DNROUTED:
			strcpy(sproto, "dnrouted");
			break;
		case RTPROT_XORP:
			strcpy(sproto, "xorp");
			break;
		case RTPROT_NTK:
			strcpy(sproto, "ntk");
			break;
		case RTPROT_DHCP:
			strcpy(sproto, "dhcp");
			break;
		default:
			sprintf(sproto, "%d", nproto);
	}
}

/*
 * convert a route table id from number to string
 */
void rt_table_n2s(unsigned ntable, char *stable)
{
	switch(ntable) {
		case RT_TABLE_UNSPEC:
			strcpy(stable, "none");
			break;
		case RT_TABLE_COMPAT:
			strcpy(stable, "compat");
			break;
		case RT_TABLE_DEFAULT:
			strcpy(stable, "default");
			break;
		case RT_TABLE_MAIN:
			strcpy(stable, "main");
			break;
		case RT_TABLE_LOCAL:
			strcpy(stable, "local");
			break;
		default:
			sprintf(stable, "%d", ntable);
	}
}
